#!/usr/bin/env python3
import pytest

from hopla.cli.groupcmds.get_user import HabiticaUser
from hopla.hoplalib.zoo.fooddata import FoodData
from hopla.hoplalib.zoo.foodmodels import FoodStockpile, FoodStockpileBuilder
from hopla.hoplalib.zoo.zoomodels import Zoo, ZooBuilder
from hopla.hoplalib.zoo.zoofeed_algorithms import FeedPlanItem, ZooFeedAlgorithm, \
    ZookeeperFeedPlan
from tests.testutils.user_test_utils import UserTestUtil


def plan_with_single_item(food_name: str, pet_name: str, times: int) -> ZookeeperFeedPlan:
    plan = ZookeeperFeedPlan()
    plan.add_to_feed_plan(pet_name=pet_name, food_name=food_name, times=times)
    return plan


class TestZookeeperFeedPlan:

    def test___init___ok(self):
        plan = ZookeeperFeedPlan()

        assert plan.feed_plan == []

    def test__repr__empty(self):
        plan = ZookeeperFeedPlan()
        result: str = repr(plan)
        assert result == "ZookeeperFeedPlan(__feed_plan=[])"

    def test__repr__nonempty(self):
        pet_name = "Wolf-White"
        food_name = "Milk"
        times = 2
        plan: ZookeeperFeedPlan = plan_with_single_item(food_name, pet_name, times)

        result: str = repr(plan)

        assert f"{pet_name=}" in result
        assert f"{food_name=}" in result
        assert f"{times=}" in result

    @pytest.mark.parametrize("plan,expected_empty", [
        (plan_with_single_item("Wolf-Red", "Fish", 1), False),
        (ZookeeperFeedPlan(), True)
    ])
    def test__is_empty(self, plan: ZookeeperFeedPlan, expected_empty: bool):
        assert plan.isempty() is expected_empty

    @pytest.mark.parametrize("plan,expected_len", [
        (plan_with_single_item("Wolf-Red", "Strawberry", 3), 1),
        (ZookeeperFeedPlan(), 0)
    ])
    def test__len(self, plan: ZookeeperFeedPlan, expected_len: int):
        assert len(plan) == expected_len

    def test_format_plan_empty(self):
        result: str = ZookeeperFeedPlan().format_plan()
        assert result == ""

    def test_format_plan_non_empty(self):
        pet_name = "Rat-Shade"
        food_name = "Chocolate"
        times = 8
        plan: ZookeeperFeedPlan = plan_with_single_item(
            pet_name=pet_name, food_name=food_name, times=times
        )

        result: str = plan.format_plan()
        assert result == f'Pet {pet_name} will get {times} {food_name}.'

    def test_add_to_feed_plan_ok(self):
        plan = ZookeeperFeedPlan()
        pet_name = "Axolotl-Zombie"
        food_name = "RottenMeat"
        times = 9
        plan.add_to_feed_plan(pet_name=pet_name,
                              food_name=food_name,
                              times=times)

        expected_plan = [
            FeedPlanItem(pet_name=pet_name, food_name=food_name, times=times)
        ]
        assert plan.feed_plan == expected_plan


class TestZooFeedAlgorithm:

    def test__repr__(self, feedable_pets_zoo: Zoo,
                     filled_stockpile: FoodStockpile):
        algorithm = ZooFeedAlgorithm(zoo=feedable_pets_zoo, stockpile=filled_stockpile)

        repr_before: str = repr(algorithm)
        plan: ZookeeperFeedPlan = algorithm.make_plan()
        repr_after: str = repr(algorithm)

        before_expected = "ZooFeedAlgorithm(\n  __zookeeper_plan=\n)"
        assert repr_before == before_expected

        after_expected = f"ZooFeedAlgorithm(\n  __zookeeper_plan={plan.format_plan()}\n)"
        assert repr_after == after_expected

    def test_make_plan_empty_stockpile_results_in_empty_plan_ok(self,
                                                                empty_zoo,
                                                                empty_stockpile):
        algorithm = ZooFeedAlgorithm(zoo=empty_zoo,
                                     stockpile=empty_stockpile)

        plan: ZookeeperFeedPlan = algorithm.make_plan()

        expected_empty_plan = ZookeeperFeedPlan()
        assert plan == expected_empty_plan
        assert algorithm.stockpile == empty_stockpile

    def test_make_plan_empty_stockpile_filled_zoo_results_in_empty_plan_ok(self,
                                                                           feedable_pets_zoo,
                                                                           empty_stockpile):
        algorithm = ZooFeedAlgorithm(zoo=feedable_pets_zoo,
                                     stockpile=empty_stockpile)

        plan: ZookeeperFeedPlan = algorithm.make_plan()

        expected_empty_plan = ZookeeperFeedPlan()
        assert plan == expected_empty_plan
        assert algorithm.stockpile == empty_stockpile

    def test_make_plan_filled_stockpile_empty_zoo_results_in_empty_plan_ok(self,
                                                                           empty_zoo,
                                                                           filled_stockpile):
        algorithm = ZooFeedAlgorithm(zoo=empty_zoo,
                                     stockpile=filled_stockpile)

        plan: ZookeeperFeedPlan = algorithm.make_plan()

        expected_empty_plan = ZookeeperFeedPlan()
        assert plan == expected_empty_plan

    def test_make_plan_filled_stockpile_hungry_zoo_results_in_plan_ok(self):
        # Wolf-Shade can be fed 9 times
        pet1, food1, food1_amount = "Wolf-Shade", "Chocolate", 10
        # Alligator-Golden can be fed 8 times
        pet2, food2, food2_amount = "Alligator-Golden", "Honey", 10
        # Wolf-Red can be fed 5 times
        pet3, food3, food3_amount = "Wolf-Red", "Strawberry", 10
        # TigerCub-Fluorite takes whatever food is most abundant
        pet4, food4, food4_amount = "TigerCub-Fluorite", "Fish", 20  # most abundant
        unused_food, unused_food_amount = "CottonCandyBlue", 5

        hungry_pets = {pet1: 5, pet2: 10, pet3: 25, pet4: 49}
        zoo: Zoo = ZooBuilder(UserTestUtil.user_with_zoo(pets=hungry_pets)).build()

        start_stockpile = (
            FoodStockpileBuilder.empty_stockpile().add_food_dict({
                food1: food1_amount, food2: food2_amount, food3: food3_amount,
                food4: food4_amount, unused_food: unused_food_amount,
            })
        )

        algorithm = ZooFeedAlgorithm(zoo=zoo, stockpile=start_stockpile)

        result_plan: ZookeeperFeedPlan = algorithm.make_plan()

        expected_first_pet_feed_times = 9
        expected_second_pet_feed_times = 8
        expected_third_pet_feed_times = 5
        expected_fourth_pet_feed_times = 1
        expected_stockpile = (
            FoodStockpileBuilder.empty_stockpile().add_food_dict({
                food1: food1_amount - expected_first_pet_feed_times,
                food2: food2_amount - expected_second_pet_feed_times,
                food3: food3_amount - expected_third_pet_feed_times,
                unused_food: unused_food_amount,
                food4: food4_amount - expected_fourth_pet_feed_times
            })
        )
        expected_feed_plan = [
            FeedPlanItem(pet1, food1, expected_first_pet_feed_times),
            FeedPlanItem(pet2, food2, expected_second_pet_feed_times),
            FeedPlanItem(pet3, food3, expected_third_pet_feed_times),
            FeedPlanItem(pet4, food4, expected_fourth_pet_feed_times)
        ]

        assert algorithm.stockpile == expected_stockpile
        assert all(feed_item in result_plan for feed_item in expected_feed_plan)
        assert result_plan == algorithm.zookeeper_plan  # sanity check

    def test_make_plan_feed_single_pet(self):
        pet_name = "Velociraptor-Skeleton"
        start_fish = 20
        food_name = "Fish"
        user = HabiticaUser({"items": {
            "pets": {pet_name: 10},
            "mounts": {},
            "food": {food_name: start_fish}
        }})
        zoo: Zoo = ZooBuilder(user).build()
        stockpile: FoodStockpile = FoodStockpileBuilder().user(user).build()
        algorithm = ZooFeedAlgorithm(zoo=zoo, stockpile=stockpile)

        feed_plan: ZookeeperFeedPlan = algorithm.make_plan()

        expected_fish_given = 8
        assert len(feed_plan) == 1
        assert feed_plan.feed_plan[0] == FeedPlanItem(
            pet_name=pet_name,
            food_name=food_name,
            times=expected_fish_given
        )

        expected_fish_leftover: int = start_fish - expected_fish_given
        actual_fish_leftover = algorithm.stockpile.as_dict()[food_name]
        assert actual_fish_leftover == expected_fish_leftover

    @pytest.fixture
    def empty_stockpile(self) -> FoodStockpile:
        return FoodStockpileBuilder().build()

    @pytest.fixture
    def empty_zoo(self) -> Zoo:
        empty_zoo_user: HabiticaUser = UserTestUtil.user_with_zoo(pets=None, mounts=None)
        return ZooBuilder(empty_zoo_user).build()

    @pytest.fixture
    def feedable_pets_zoo(self) -> Zoo:
        hungry_pets = {"Parrot-Shade": 5, "Alligator-Golden": 10,
                       "Dragon-CottonCandyBlue": 20, "Wolf-Red": 25}
        user_with_feedable_pets: HabiticaUser = UserTestUtil.user_with_zoo(pets=hungry_pets)
        return ZooBuilder(user_with_feedable_pets).build()

    @pytest.fixture
    def filled_stockpile(self) -> FoodStockpile:
        stockpile = FoodStockpileBuilder().build()
        for food_name in FoodData.drop_food_names:
            lots_of_food = 100
            stockpile.add_food(food_name, n=lots_of_food)
        return stockpile



#!/usr/bin/env python3
"""
The module with CLI code that handles the `hopla hatch-all` command.
"""
import sys
from typing import Dict, List

import click

from hopla.hoplalib.hatchery.eggmodels import EggCollection
from hopla.hoplalib.hatchery.hatchalgorithms import HatchPlan, HatchPlanMaker
from hopla.hoplalib.hatchery.hatchcontroller import HatchRequester
from hopla.hoplalib.hatchery.hatchpotionmodels import HatchPotionCollection
from hopla.hoplalib.throttling import ApiRequestThrottler
from hopla.hoplalib.user.usercontroller import HabiticaUserRequest
from hopla.hoplalib.user.usermodels import HabiticaUser
from hopla.hoplalib.zoo.foodmodels import FeedStatus
from hopla.hoplalib.zoo.petmodels import Pet


@click.command()
def hatch_all():
    """Hatch all the available eggs.

    \b
    hopla hatch-all - Print the possible hatchings, and, if the user
                      confirms, hatch all the available eggs
                      with all the available hatching potions.

    \b
    Examples:
    ----
    $ hopla hatch-all
    > A Treeling egg will be hatched by a Zombie potion.
    > A Robot egg will be hatched by a Desert potion.
    > Do you wish to proceed? [y/N]: yes
    Successfully hatched a Treeling-Zombie.
    Successfully hatched a Robot-Desert.

    """
    user: HabiticaUser = HabiticaUserRequest().request_user_data_or_exit()
    eggs = EggCollection(user.get_eggs())
    potions = HatchPotionCollection(user.get_hatch_potions())
    pets: List[Pet] = to_pet_list(user.get_pets())

    plan_maker = HatchPlanMaker(
        egg_collection=eggs, hatch_potion_collection=potions, pets=pets
    )
    plan: HatchPlan = plan_maker.make_plan()

    if plan.is_empty():
        click.echo("The hatch plan is empty. Do you have enough pets and hatching potions?")
        click.echo("Exiting")
        sys.exit(1)

    plan_text: str = plan.format_plan()
    user_confirmed: bool = click.confirm(text=plan_text + "Do you wish to proceed?")
    if user_confirmed is True:
        _hatch_eggs(plan)
    else:
        click.echo("No eggs were hatched.")


def _hatch_eggs(plan: HatchPlan) -> None:
    """Hatch all the eggs. Print the result to the terminal.
    Warning: this function does not ask for confirmation.
    """
    throttler = ApiRequestThrottler(api_requests=[
        HatchRequester(item.egg.name, item.potion.name).post_hatch_egg for item in plan
    ])

    for hatch_request in throttler.release():
        response_msg: str = hatch_request()
        click.echo(response_msg)


def to_pet_list(pets: Dict[str, int]) -> List[Pet]:
    """Helper method that takes a pet_dict and returns a List[Pet]."""
    return [
        Pet(name, feed_status=FeedStatus(n)) for (name, n) in pets.items()
    ]



#!/usr/bin/env python3
"""
Module with tasks models.
"""
import datetime
from typing import Dict, List, Optional


class TaskDifficultyData:
    """Class with task difficulty data."""
    DIFFICULTIES_SCORE_MAPPING: Dict[str, str] = {
        "hard": "2", "medium": "1.5", "easy": "1", "trivial": "0.1"
    }
    VALID_DIFFICULTIES: List[str] = list(DIFFICULTIES_SCORE_MAPPING.keys())
    """Habitica task difficulties. See https://habitica.com/apidoc/#api-Task-CreateUserTasks."""


class HabiticaChecklist:
    """ Habitica Checklist """

    def __init__(self, *, checklist: Optional[List[str]] = None):
        self.checklist = checklist or []

    def __repr__(self) -> object:
        return self.__class__.__name__ + f"(checklist={self.checklist})"

    def is_empty(self) -> bool:
        """Return true if the checklist is empty.

        >>> HabiticaChecklist(checklist=[]).is_empty()
        True
        >>> HabiticaChecklist(checklist=["item 1"]).is_empty()
        False

        """
        return self.checklist is None or len(self.checklist) == 0

    def to_json_list(self) -> List[Dict[str, str]]:
        """Turn a checklist into a list that we can pass to the habitica API"""
        return [{"text": line} for line in self.checklist]


class HabiticaTodo:
    """ Habitica To-Do"""

    def __init__(self, *,
                 todo_name: str,
                 difficulty: str,
                 due_date: Optional[datetime.datetime] = None,
                 checklist: Optional[HabiticaChecklist] = None):
        self.todo_name = todo_name
        self._type = "todo"
        self.difficulty = difficulty
        self.due_date = due_date
        self.checklist: Optional[HabiticaChecklist] = checklist or HabiticaChecklist()

    def difficulty_to_score(self) -> str:
        """Return score for the To-Do's difficulty.

        >>> hab_todo = HabiticaTodo(todo_name="my-task", difficulty="easy")
        >>> hab_todo.difficulty_to_score()
        '1'
        >>> hab_todo.difficulty = "trivial"
        >>> hab_todo.difficulty_to_score()
        '0.1'

        :raise KeyError in case of an invalid difficulty
        """
        return TaskDifficultyData.DIFFICULTIES_SCORE_MAPPING[self.difficulty]

    def due_date_to_date_str(self) -> str:
        """Turn the due_date into a valid ISO 8601 date format

        >>> hab_todo = HabiticaTodo(todo_name="my-task", difficulty="easy",\
                                    due_date=datetime.datetime(day=31, month=1, year=2069))
        >>> hab_todo.due_date_to_date_str()
        '2069-01-31'
        """
        habitica_date_format = "%Y-%m-%d"
        return self.due_date.strftime(habitica_date_format)

    def to_json_dict(self) -> dict:
        """Turn an instance of this class into a dict that we can pass to the habitica API.

        >>> hab_todo = HabiticaTodo(todo_name="my-task", difficulty="medium",\
                                    checklist=HabiticaChecklist(checklist=["myitem"]))
        >>> hab_todo.to_json_dict()
        {'text': 'my-task', 'type': 'todo', 'priority': '1.5', 'checklist': [{'text': 'myitem'}]}
        """
        todo_dict = {
            "text": self.todo_name,
            "type": self._type,
            "priority": self.difficulty_to_score(),
            "checklist": self.checklist.to_json_list()
        }

        if self.due_date:
            todo_dict["date"] = self.due_date_to_date_str()

        return todo_dict



#!/usr/bin/env python3
"""
Module with models for hatching potions.
"""
from dataclasses import dataclass, field
from typing import Dict

from hopla.hoplalib.errors import YouFoundABugRewardError
from hopla.hoplalib.hatchery.hatchdata import HatchPotionData


class HatchPotionException(YouFoundABugRewardError):
    """Exception raised when there is an error with an hatching potion."""


@dataclass
class HatchPotion:
    """A habitica hatching potion."""
    name: str
    quantity: int = 1

    def __post_init__(self):
        if self.name not in HatchPotionData.hatch_potion_names:
            raise HatchPotionException(f"{self.name} is not a valid hatching potion name.")
        if self.quantity < 0:
            raise HatchPotionException(f"{self.quantity} is below 0.")

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self.name}: {self.quantity})"

    def is_standard_hatch_potion(self) -> bool:
        """Return true if this is a standard hatching potion."""
        return self.name in HatchPotionData.drop_hatch_potion_names

    def is_magic_hatch_potion(self) -> bool:
        """Return true if this is a magic hatching potion."""
        return self.name in HatchPotionData.magic_hatch_potion_names

    def is_wacky_hatch_potion(self) -> bool:
        """Return true if this is a wacky hatching potion."""
        return self.name in HatchPotionData.wacky_hatch_potion_names


@dataclass
class HatchPotionCollection:
    """A collection of hatching potions, backed by a Dict[str, Egg]."""

    potions: Dict[str, int] = field(init=True, compare=False, default_factory=dict)
    __potions: Dict[str, HatchPotion] = field(init=False)

    def __post_init__(self):
        """Use the given potions to create __potions."""
        self.__potions: Dict[str, HatchPotion] = {
            name: HatchPotion(name, quantity=quantity)
            for (name, quantity) in self.potions.items()
        }

    def __len__(self) -> int:
        return len(self.__potions)

    def __iter__(self):
        return iter(self.__potions)

    def __getitem__(self, name: str) -> HatchPotion:
        return self.__potions[name]

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self.__potions=})"

    def values(self):
        """Like dict.values but for an HatchPotionCollection."""
        for potion in self.__potions.values():
            yield potion

    def remove_hatch_potion(self, potion: HatchPotion) -> "HatchPotionCollection":
        """Remove a single hatching potion from this collection.

        :param potion: The potion to remove.
        :return: The updated HatchPotionCollection
        """
        if self.__potions.get(potion.name) is None:
            raise HatchPotionException(
                f"{potion.name} was not in the collection {self.__potions}"
            )
        if self.__potions.get(potion.name).quantity == 0:
            raise HatchPotionException(
                f"We had 0 {potion.name} in the collection. Cannot remove any more."
            )

        self.__potions[potion.name].quantity -= 1
        return self



"""
The module with CLI code that handles the `hopla complete` command.
"""

import abc
import logging
import os
from pathlib import Path

import click
from hopla.hoplalib.errors import PrintableException

log = logging.getLogger()


class RecognizedShell(abc.ABC):
    """
    Shells that Hopla recognizes for the 'hopla complete [shell]' command
    """

    def __init__(self, shell_name: str):
        self.shell_name = shell_name

    def __str__(self):
        return self.__class__.__name__ + "(" + f"shell_name={self.shell_name}" + ")"

    @abc.abstractmethod
    def get_generated_autocomplete_cmd(self) -> str:
        """Returns the command needed to generate autocomplete code"""

    @abc.abstractmethod
    def supports_automatic_complete_enablement(self) -> bool:
        """Returns whether this shell supports automatic enablement of autocomplete """

    @abc.abstractmethod
    def handle_enable_complete_request(self):
        """Handle request for enabling autocomplete regardless if there is support for it"""


class RecognizedShellWithAutomaticAutocompleteDisabled(RecognizedShell,
                                                       abc.ABC):
    """
    Shells that Hopla currently does not have `hopla complete [shell] --enable`
    support for.
    """

    @abc.abstractmethod
    def get_manual_autocomplete_instructions(self) -> [str]:
        """Returns instructions on how the enable hopla autocompletion for this shell"""

    def handle_enable_complete_request(self) -> [str]:
        return self.get_manual_autocomplete_instructions()

    def supports_automatic_complete_enablement(self) -> bool:
        return False


class RecognizedShellWithAutomaticAutocompleteEnabled(RecognizedShell,
                                                      abc.ABC):
    """
    Shells that Hopla currently has hopla complete [shell] --enable support for
    """

    @abc.abstractmethod
    def enable_autocomplete(self) -> bool:
        """ Enables hopla autocompletion, """

    @abc.abstractmethod
    def autocomplete_is_enabled(self) -> bool:
        """ Return True if autocompletion is already enabled"""

    def handle_enable_complete_request(self) -> None:
        self.enable_autocomplete()

    def supports_automatic_complete_enablement(self) -> bool:
        return True


class AlreadyEnabledException(PrintableException):
    """Raised when autocomplete is already enabled"""


class Bash(RecognizedShellWithAutomaticAutocompleteEnabled):
    """
    Class representing the
    [Bash Shell](https://en.wikipedia.org/wiki/Bash_(Unix_shell))
    """

    def __init__(self):
        super().__init__("bash")
        self.config_file: Path = Path.home() / ".bashrc"

    def get_generated_autocomplete_cmd(self) -> str:
        return f"_HOPLA_COMPLETE={self.shell_name}_source hopla"

    def get_enablement_code_for_bashrc(self):
        """Return the code that is used to enable autocompletion"""
        return f'eval "$({self.get_generated_autocomplete_cmd()})"'

    def autocomplete_is_enabled(self) -> bool:
        """Returns true if the autocompletion enablement code is in ~/.bashrc"""
        with open(self.config_file, mode="r", encoding="utf-8") as bashrc:
            enablement_code = self.get_enablement_code_for_bashrc()
            for bashrc_line in bashrc.readlines():
                if enablement_code in bashrc_line:
                    return True
        return False

    def enable_autocomplete(self):
        enablement_code_for_bashrc = self.get_enablement_code_for_bashrc()

        if self.autocomplete_is_enabled():
            msg = f"{enablement_code_for_bashrc} was already in {self.config_file}"
            raise AlreadyEnabledException(msg)

        with open(self.config_file, mode="a", encoding="utf-8") as bashrc:
            bashrc.writelines([enablement_code_for_bashrc, "\n"])


class Zsh(RecognizedShellWithAutomaticAutocompleteDisabled):
    """
    Class representing the [Z shell](https://en.wikipedia.org/wiki/Z_shell)
    """

    def __init__(self):
        super().__init__("zsh")
        self.example_zsh_complete_file: str = "~/hopla-zsh-completion.zsh"
        self.config_file: str = "~/.zshrc"

    def get_generated_autocomplete_cmd(self) -> str:
        return f"_HOPLA_COMPLETE={self.shell_name}_source hopla"

    def get_manual_autocomplete_instructions(self) -> [str]:
        return [
            f"   hopla autocomplete {self.shell_name} > {self.example_zsh_complete_file}",
            f"Then source the file in your {self.config_file}",
            f"   . {self.example_zsh_complete_file}"
        ]


class Fish(RecognizedShellWithAutomaticAutocompleteDisabled):
    """
    Class representing the [Fish shell](https://en.wikipedia.org/wiki/Fish_(Unix_shell))
    """

    def __init__(self):
        super().__init__("fish")
        self.hopla_complete_file = "~/.config/fish/completions/hopla.fish"

    def get_generated_autocomplete_cmd(self) -> str:
        return f"env _HOPLA_COMPLETE={self.shell_name}_source hopla"

    def get_manual_autocomplete_instructions(self) -> [str]:
        return [f"   hopla autocomplete {self.shell_name} > {self.hopla_complete_file}"]


supported_shell_mapping = {"bash": Bash, "zsh": Zsh, "fish": Fish}
supported_shell_id_strings = list(supported_shell_mapping.keys())


def get_shell_by_name(shell_name: str) -> RecognizedShell:
    """Get a shell by its name"""
    shell_class = supported_shell_mapping.get(shell_name)
    if shell_class is None:
        # defensive programming: should be handled by click.Choice already
        raise ValueError(f"Unexpected shell {shell_name}")

    return shell_class()


autocomplete_supported_shells = click.Choice(supported_shell_id_strings)


@click.command()
@click.argument("shell", type=autocomplete_supported_shells, default="bash")
@click.option("--enable/--no-enable", is_flag=True, default=False, show_default=True,
              help="When True, automatically add hopla autocompletion to your shell's config. "
                   "When False, just print the autocomplete code to the terminal")
def complete(shell: str, enable: bool):
    """Print or enable shell autocompletion.

    \b
    Examples
    ---
    # automatically add autocompletion for bash (bash is used when no shell is provided)
    $ hopla complete --enable

    \b
    # show the autocomplete code for zsh
    $ hopla complete zsh


    \f
    :param shell:
    :param enable:
    :return:
    """
    log.debug(f"hopla complete {shell} --enable={enable}")

    shell_obj: RecognizedShell = get_shell_by_name(shell_name=shell)

    if enable is False:
        show_autocomplete_code(shell=shell_obj)
    else:
        try_enable_autocomplete(shell=shell_obj)


def try_enable_autocomplete(shell: RecognizedShell):
    """Try to enable autocompletion. If it is already enabled, inform the user.
    If it cannot be enabled automatically, provide instructions"""
    if shell.supports_automatic_complete_enablement():
        try:
            enable_autocomplete(shell)
        except AlreadyEnabledException as ex:
            click.echo(ex.msg)
    else:
        show_manual_autocomplete_code(shell=shell)


def show_manual_autocomplete_code(shell: RecognizedShell):
    """Print instructions on how to enable autocomplete for the specified shell if
    --enable is not supported by hopla.

    :param shell:
    :return:
    """
    click.echo(f"Sorry, automatic installation is not supported for {shell.shell_name}.")
    click.echo("You can enable autocompletion manually by running:")
    instructions: [str] = shell.handle_enable_complete_request()
    for instruction in instructions:
        click.echo(instruction)


def enable_autocomplete(shell: RecognizedShell):
    """
    Adds a command to the relevant configuration files s.t. autocompletion
    code is automatically loaded. This is used with the `--enable` flag for
    hopla complete [shell] --enable commands.

    :param shell:
    :return:
    """
    shell.handle_enable_complete_request()
    click.echo("enabled autocompletion")
    click.echo(f"restart {shell.shell_name} to make use of it")


def show_autocomplete_code(shell: RecognizedShell):
    """
    Prints autocompletion code for the passed shell.

    :param shell:
    :return:
    """
    cmd_str = shell.get_generated_autocomplete_cmd()
    os.system(cmd_str)



#!/usr/bin/env python3
"""
Module with spell casting logic.
"""
from dataclasses import dataclass

import requests

from hopla.hoplalib.cast.spellmodel import Spell
from hopla.hoplalib.http import HabiticaRequest, UrlBuilder


@dataclass
class PostCastRequest(HabiticaRequest):
    """HabiticaRequester that POSTs to the spell cast endpoint.

    [APIDOCS](https://habitica.com/apidoc/#api-User-UserCast)
    """
    spell: Spell

    @property
    def url(self) -> str:
        """Get the url"""
        path_extension = f"/user/class/cast/{self.spell.spell_name}"
        return UrlBuilder(path_extension=path_extension).url

    def post_spell(self) -> requests.Response:
        """Perform the user get request and return the response"""
        return requests.post(url=self.url, headers=self.default_headers)



"""
The module with CLI code that handles the `hopla support-development` command.
"""
import logging

import click
import requests

from hopla.hoplalib.common import GlobalConstants
from hopla.hoplalib.http import UrlBuilder, RequestHeaders
from hopla.hoplalib.requests_helper import get_data_or_exit

log = logging.getLogger()


@click.command()
@click.option("--gems", "-g", type=click.IntRange(min=0), default=4, show_default=True,
              metavar="N_GEMS", help="N_GEMS is the number of gems you wish to contribute")
def support_development(gems: int):
    """Support the development of Hopla.

    Without options, hopla support-development sends 4 gems to the
    development to be used for testing new features. Use the --gems option
    to change the number of gems send.

    Thanks for supporting Hopla! Your donation is thoroughly appreciated =D.


    \b
    Examples
    -----
    # support development by donating 4 gems
    hopla support-development

    # support development by donating 20 gems
    hopla support-development -g 20


    [APIdocs](https://habitica.com/apidoc/#api-Member-TransferGems)

    \f
    :param gems:
    :return:
    """
    log.debug(f"hopla support-development gems={gems}")
    url = UrlBuilder(path_extension="/members/transfer-gems").url
    headers = RequestHeaders().get_default_request_headers()

    params = {
        "message": "Thanks!",
        "toUserId": GlobalConstants.DEVELOPMENT_UUID,
        "gemAmount": gems
    }

    support_development_request = requests.Request(
        method="POST", url=url, headers=headers, json=params
    )
    response: requests.Response = requests.session().send(support_development_request.prepare())
    response_data = get_data_or_exit(response)

    click.echo(response_data)
    click.echo()
    click.echo("Thanks for supporting Hopla! Your donation is thoroughly appreciated =D.")



#!/usr/bin/env python3
from typing import Any, Callable, List

import pytest

from hopla.hoplalib.throttling import ApiRequestThrottler


class TestApiRequestThrottler:

    def test__init__(self, ten_api_calls: List[Callable]):
        secs = 3
        call_limit = 12
        throttler = ApiRequestThrottler(ten_api_calls,
                                        throttle_seconds=secs,
                                        throttle_limit=call_limit)

        assert throttler.api_requests == ten_api_calls
        assert throttler.total_calls == len(ten_api_calls)
        assert throttler.throttle_limit == call_limit
        assert throttler.throttle_seconds == 3

    def test_release_without_throttle(self, two_api_requests: List[Callable]):
        throttler = ApiRequestThrottler(two_api_requests)

        call_releaser_result = throttler.release()
        first: Callable = call_releaser_result.__next__()
        second: Callable = call_releaser_result.__next__()
        with pytest.raises(StopIteration):
            call_releaser_result.__next__()

        assert first() == two_api_requests[0]()
        assert second() == two_api_requests[1]()

    def test_execute_all_request_with_throttle(self,
                                               three_api_requests: List[Callable],
                                               capsys):
        secs = 0.001
        throttler = ApiRequestThrottler(three_api_requests,
                                        throttle_limit=1,
                                        throttle_seconds=secs)

        throttler.execute_all_requests()

        captured = capsys.readouterr()
        expected_stderr_msg = \
            f"To prevent overheating Habitica, we'll wait {secs} seconds every request.\n"
        assert captured.err == expected_stderr_msg
        assert captured.out == "some api output"

    @pytest.mark.parametrize(
        "api_request,throttle_limit,expected_should_throttle", [
            ([], 10, False),
            ([lambda: 1, lambda: 2], 1, True),
            ([lambda: 1, lambda: 2], 2, False),
            ([lambda: 1, lambda: 2, lambda: 3], 3, False),
            ([lambda: "a"] * 12, 9, True),
            ([lambda: "a"] * 12, 17, False)
        ]
    )
    def test_exceeds_throttle_limit_noargs(self, api_request: List[Callable],
                                           throttle_limit: int,
                                           expected_should_throttle: bool):
        throttler = ApiRequestThrottler(api_request, throttle_limit=throttle_limit)

        result: bool = throttler.exceeds_throttle_limit()

        assert result == expected_should_throttle

    @pytest.fixture
    def two_api_requests(self) -> List[Callable[[None], Any]]:
        def f1() -> bool:
            return "Thermometer".islower()

        def f2() -> str:
            return "Nebuliser".upper()

        return [f1, f2]

    @pytest.fixture
    def three_api_requests(self, two_api_requests) -> List[Callable]:
        def f3():
            print("some api output", end="")

        return two_api_requests + [f3]

    @pytest.fixture
    def ten_api_calls(self) -> List[Callable[[None], None]]:
        def api_call():
            print("hi")

        return [api_call] * 10



#!/usr/bin/env python3

import pytest
from click.testing import CliRunner, Result
from unittest.mock import patch, MagicMock
from requests.status_codes import codes

from hopla.cli.feed_all import feed_all
from hopla.cli.groupcmds.get_user import HabiticaUser


class MockBadGatewayResponse:
    def __init__(self, msg):
        self.status_code = codes.bad_gateway
        self.__json = {"success": False, "error": "BadGateway", "message": msg}

    def json(self):
        return self.__json


class MockOkResponse:
    def __init__(self, msg: str):
        self.status_code = codes.ok
        self.__json = {"success": True, "data": -1, "message": msg}

    def json(self):
        return self.__json


class TestFeedAllCliCommand:
    yes_responses = ["y", "Y", "yes", "Yes", "YES"]
    no_responses = ["n", "N", "no", "No", "NO", "", "anything else"]
    released_zoo_users = [
        HabiticaUser({
            "items": {"pets": {"Seahorse-White": 0},
                      "mounts": {"Seahorse-White": None},
                      "food": {"Milk": 20}}
        }),
        HabiticaUser({
            "items": {"pets": {"Egg-White": 0},
                      "mounts": {"Egg-White": True},
                      "food": {"Milk": 10}}
        })
    ]

    @pytest.mark.parametrize("no_response", no_responses)
    @patch("hopla.cli.feed_all.HabiticaUserRequest.request_user_data_or_exit")
    def test_feed_all_pets_user_aborts(self,
                                       mock_user_request: MagicMock,
                                       no_response: str,
                                       user_with_feedable_pet: HabiticaUser):
        mock_user_request.return_value = user_with_feedable_pet
        runner = CliRunner()
        result: Result = runner.invoke(feed_all, input=no_response)

        assert "Aborted!" in result.stdout
        assert result.exit_code == 1

    @patch("hopla.cli.feed_all.HabiticaUserRequest.request_user")
    def test_feed_all_api_bad_gateway_fail(self, mock_user_request: MagicMock):
        bad_gateway_msg = "Service responded with a status of 502 (Bad Gateway)"
        mock_user_request.return_value = MockBadGatewayResponse(msg=bad_gateway_msg)

        runner = CliRunner()
        result: Result = runner.invoke(feed_all, input="Yes")

        assert result.exit_code == 1
        assert bad_gateway_msg in result.stdout

    @patch("hopla.cli.feed_all.HabiticaUserRequest.request_user_data_or_exit")
    def test_feed_all_empty_plan_ok(self,
                                    mock_user_request: MagicMock,
                                    user_with_no_feedable_pets: HabiticaUser):
        mock_user_request.return_value = user_with_no_feedable_pets

        runner = CliRunner()
        result: Result = runner.invoke(feed_all, input="Yes")

        assert result.exit_code == 0
        assert result.stdout.startswith("The feed plan is empty.")

    @pytest.mark.parametrize("yes_response", yes_responses)
    @patch("hopla.cli.feed_all.FeedPostRequester.post_feed_request")
    @patch("hopla.cli.feed_all.HabiticaUserRequest.request_user_data_or_exit")
    def test_feed_all_ok(self,
                         mock_user_request: MagicMock,
                         mock_feed_request: MagicMock,
                         user_with_feedable_pet: HabiticaUser,
                         yes_response: str):
        # This user has a pet that can be grown into a mount by feeding 8 Fish
        mock_user_request.return_value = user_with_feedable_pet

        feed_msg = "You have tamed Skeleton Velociraptor, let's go for a ride!"
        mock_feed_request.return_value = MockOkResponse(msg=feed_msg)

        runner = CliRunner()
        result: Result = runner.invoke(feed_all, input=yes_response)

        assert "Pet Velociraptor-Skeleton will get 8 Fish.\n" in result.stdout
        assert f"Do you want to proceed? [y/N]: {yes_response}\n" in result.stdout
        assert feed_msg in result.stdout
        assert '"feed_status": -1' in result.stdout
        assert result.exit_code == 0

    @pytest.mark.parametrize("yes_response", yes_responses)
    @pytest.mark.parametrize("released_zoo_user", released_zoo_users)
    @patch("hopla.cli.feed.HabiticaUserRequest.request_user_data_or_exit")
    def test_feed_released_pets(self, mock_user_request: MagicMock,
                                released_zoo_user: HabiticaUser,
                                yes_response: str):
        mock_user_request.return_value = released_zoo_user

        runner = CliRunner()
        result: Result = runner.invoke(feed_all, input=yes_response)

        expected_msg = ("The feed plan is empty. Reasons could include:\n"
                        "1. There is insufficient food available to turn pets into mounts\n"
                        "2. You don't have any feedable pets.")
        assert expected_msg in result.stdout

    @pytest.fixture
    def user_with_feedable_pet(self) -> HabiticaUser:
        return HabiticaUser({"items": {
            "pets": {"Velociraptor-Skeleton": 10},
            "mounts": {},
            "food": {"Fish": 20}
        }})

    @pytest.fixture
    def user_with_no_feedable_pets(self) -> HabiticaUser:
        return HabiticaUser({"items": {
            "pets": {},
            "mounts": {},
            "food": {"Fish": 20}
        }})



#!/usr/bin/env python3
from hopla.hoplalib.user.usermodels import HabiticaUser


class TestHabiticaUser:

    def test_get_stats(self):
        user_test_stat_values = {
            "buffs": {
                "str": 50, "int": 50, "per": 3206, "con": 50, "stealth": 0, "streaks": False,
                "snowball": False, "spookySparkles": False, "shinySeed": False, "seafoam": False
            },
            "training": {"int": 0, "per": 0, "str": 0, "con": 0},
            "hp": 50, "mp": 65.7, "exp": 2501, "gp": 1072327.9, "lvl": 121,
            "class": "wizard", "points": 0, "str": 0, "con": 0, "int": 12, "per": 88,
            "toNextLevel": 5010, "maxHealth": 50, "maxMP": 304
        }

        user = HabiticaUser(user_dict={"stats": user_test_stat_values})

        assert user.get_stats() == user_test_stat_values

    def test_get_auth(self):
        user_test_auth_values = {
            "local": {"username": "hopla", "lowerCaseUsername": "hopla",
                      "email": "something+habitica@gmail.com"
                      },
            "timestamps": {"created": "2022-03-22T24:23:38.119Z",
                           "loggedin": "2022-09-18T08:47:45.286Z",
                           "updated": "2022-09-18T14:20:55.530Z"
                           },
            "facebook": {}, "google": {}, "apple": {}
        }

        user = HabiticaUser(user_dict={"auth": user_test_auth_values})

        assert user.get_auth() == user_test_auth_values

    def test_get_inventory(self):
        inventory = {
            "gear": {"equipped": {"back": "back_special_aetherCloak"},
                     "costume": {"armor": "armor_armoire_bluePartyDress", "body": "body_base_0"},
                     "owned": {"armor_special_fall2019Healer": True}},
            "special": {"goodluck": 9000},
            "lastDrop": {"count": 80, "date": "2021-10-12T15:45:30.384Z"},
            "pets": {"Cactus-Golden": 0, "Unicorn-Red": -1, "Wolf-CottonCandyPink": 5},
            "eggs": {"Dragon": 338, "Nudibranch": 3, "TRex": 0},
            "hatchingPotions": {"Desert": 456, "MossyStone": 1},
            "food": {"RottenMeat": 846},
            "mounts": {"Fox-RoyalPurple": True, "Dragon-Skeleton": None, "Wolf-MossyStone": True},
            "quests": {"trex_undead": 0},
            "currentPet": "Egg-Base",
            "currentMount": "Aether-Invisible"
        }
        user = HabiticaUser(user_dict={"items": inventory})
        assert user.get_inventory() == inventory

    def test_get_gp(self):
        gp = 12.0
        user = HabiticaUser(user_dict={"stats": {"gp": gp}})
        assert user.get_gp() == gp

    def test_get_mp(self):
        mp = 112.0
        user = HabiticaUser(user_dict={"stats": {"mp": mp}})
        assert user.get_mp() == mp

    def test_get_pets(self):
        pets = {"Spider-Base": -1, "TRex-Skeleton": 5}
        user = HabiticaUser(user_dict={"items": {"pets": pets}})
        assert user.get_pets() == pets

    def test_get_mounts(self):
        mounts = {"Spider-Base": None, "TRex-Skeleton": True}
        user = HabiticaUser(user_dict={"items": {"mounts": mounts}})
        assert user.get_mounts() == mounts

    def test_get_food(self):
        food = {"CottonCandyBlue": 10, "Fish": 830}
        user = HabiticaUser(user_dict={"items": {"food": food}})
        assert user.get_food() == food

    def test_get_hatch_potions(self):
        hatch_potions = {"Base": 10, "SolarSystem": 1009}
        user = HabiticaUser(user_dict={"items": {"hatchingPotions": hatch_potions}})
        assert user.get_hatch_potions() == hatch_potions

    def test_get_eggs(self):
        eggs = {"Fox": 1001, "Nudibranch": 9}
        user = HabiticaUser(user_dict={"items": {"eggs": eggs}})
        assert user.get_eggs() == eggs



#!/usr/bin/env python3
from hopla.cli.groupcmds.get_user import HabiticaUser
from hopla.hoplalib.zoo.foodmodels import FeedStatus
from hopla.hoplalib.zoo.petmodels import Pet, PetMountPair
from hopla.hoplalib.zoo.zoomodels import Zoo, ZooBuilder, ZooHelper
from tests.testutils.user_test_utils import UserTestUtil


class TestZooBuilder:
    def test__init__(self):
        animal_name, feed_status = "Wolf-Zombie", 5
        animal_name2, feed_status2 = "Deer-Base", 10

        pets_dict = {animal_name: feed_status, animal_name2: feed_status2}
        mounts_dict = {animal_name: True}
        user: HabiticaUser = UserTestUtil.user_with_zoo(pets=pets_dict, mounts=mounts_dict)

        builder = ZooBuilder(user)

        assert builder.pets == pets_dict
        assert builder.mounts == mounts_dict

    def test__repr__(self):
        animal_name, feed_status = "Wolf-CottonCandyBlue", 5
        pets_dict = {animal_name: feed_status}
        mounts_dict = {animal_name: True}
        user: HabiticaUser = UserTestUtil.user_with_zoo(pets=pets_dict, mounts=mounts_dict)

        builder_str = str(ZooBuilder(user))

        assert "'pets': " + str(pets_dict) in builder_str
        assert "'mounts': " + str(mounts_dict) in builder_str

    def test_build_empty_zoo(self):
        empty_zoo = {"pets": {}, "mounts": {}}
        user = HabiticaUser({"items": empty_zoo})

        zoo: Zoo = ZooBuilder(user).build()

        assert zoo == {}

    def test_build_released_zoo(self):
        pass

    def test_build_raised_pets(self):
        animal_name = "BearCub-Shadow"
        feed_status = -1  # i.e. No pet, just mount
        user: HabiticaUser = UserTestUtil.user_with_zoo(pets={animal_name: feed_status},
                                                        mounts={animal_name: True})

        result_zoo: Zoo = ZooBuilder(user).build()

        assert len(result_zoo) == 1

        result_pair: PetMountPair = result_zoo[animal_name]
        assert result_pair.pet.name == animal_name
        assert result_pair.pet.feed_status == FeedStatus(feed_status)
        assert result_pair.mount_available()
        assert result_pair.pet_available() is False

    def test_build_no_pet_yes_mount(self):
        animal_name = "Aether-Invisible"
        user: HabiticaUser = UserTestUtil.user_with_zoo(mounts={animal_name: True})

        zoo: Zoo = ZooBuilder(user).build()

        assert len(zoo) == 1
        result_pair: PetMountPair = zoo[animal_name]
        assert result_pair.pet is None
        assert result_pair.mount_available()
        assert result_pair.pet_available() is False

    def test_build_yes_pet_no_mount(self):
        animal_name = "Dragon-Skeleton"
        feed_status = 27
        user = UserTestUtil.user_with_zoo(pets={animal_name: feed_status})

        zoo: Zoo = ZooBuilder(user).build()

        assert len(zoo) == 1
        result_pair: PetMountPair = zoo[animal_name]
        assert result_pair.pet.name == animal_name
        assert result_pair.pet.feed_status == FeedStatus(feed_status)
        assert result_pair.mount_available() is False
        assert result_pair.pet_available()


class TestZooHelper:
    def test_filter_on_pet_mount_pair(self):
        pets = {"Owl-Golden": -1, "Ferret-Red": 27, "Phoenix-Base": 5, "Parrot-Base": 10}
        mounts = {"BearCub-Desert": True}
        user: HabiticaUser = UserTestUtil.user_with_zoo(pets=pets, mounts=mounts)
        zoo: Zoo = ZooBuilder(user).build()

        def predicate(pair: PetMountPair) -> bool:
            # arbitrary filter function
            return pair.pet is not None and (
                    pair.pet.name == "Parrot-Base"
                    or pair.mount_available()
                    or int(pair.pet.feed_status) == 27
            )

        helper = ZooHelper(zoo)
        filtered_zoo: Zoo = helper.filter_on_pet_mount_pairs(predicate=predicate)

        expected_pets = sorted(["Ferret-Red", "Parrot-Base"])
        assert all(expected_pet in filtered_zoo.keys() for expected_pet in expected_pets)

    def test_get_feedable_zoo(self):
        pets = {
            "BearCub-Desert": 5,  # we have the mount, not hungry
            "Owl-Golden": -1,  # we don't have the pet
            "Ferret-Red": 27,  # can be fed
            "Phoenix-Base": 5,  # unfeedable pet
            "Parrot-Base": 10,  # can be fed
        }
        mounts: dict = {"BearCub-Desert": True}
        user: HabiticaUser = UserTestUtil.user_with_zoo(pets=pets, mounts=mounts)
        zoo: Zoo = ZooBuilder(user).build()

        helper = ZooHelper(zoo)
        filtered_zoo: Zoo = helper.get_feedable_zoo()

        expected_pets = ["Ferret-Red", "Parrot-Base"]
        assert all(expected_pet in filtered_zoo.keys() for expected_pet in expected_pets)

    def test_filter_on_pet_name(self):
        pets = {
            "BearCub-Desert": 5, "Owl-Golden": -1, "Ferret-Red": 27,
            "Phoenix-Base": 5, "Parrot-Base": 10
        }
        user: HabiticaUser = UserTestUtil.user_with_zoo(pets=pets)
        zoo: Zoo = ZooBuilder(user).build()

        helper = ZooHelper(zoo)
        filtered_zoo: Zoo = helper.filter_on_pet_name(lambda name: name.endswith("-Base"))

        expected_pets = ["Phoenix-Base", "Parrot-Base"]
        assert all(expected_pet in filtered_zoo.keys() for expected_pet in expected_pets)

    def test_filter_on_pet(self):
        pets: dict = {"BearCub-Desert": 9, "Owl-Golden": -1, "Ferret-Red": 27, "Phoenix-Base": 5,
                      "Parrot-Base": 10}
        user: HabiticaUser = UserTestUtil.user_with_zoo(pets=pets)
        zoo: Zoo = ZooBuilder(user).build()

        def predicate(pet: Pet) -> bool:
            return int(pet.feed_status) < 9

        helper = ZooHelper(zoo)
        filtered_zoo: Zoo = helper.filter_on_pet(predicate=predicate)

        expected_pets = ["Owl-Golden", "Phoenix-Base"]
        assert all(expected_pet in filtered_zoo.keys() for expected_pet in expected_pets)



#!/usr/bin/env python3
"""
Module with models for eggs.
"""
from dataclasses import dataclass, field
from typing import Any, Dict

from hopla.hoplalib.errors import YouFoundABugRewardError
from hopla.hoplalib.hatchery.hatchdata import EggData
from hopla.hoplalib.hatchery.hatchpotionmodels import HatchPotion


class EggException(YouFoundABugRewardError):
    """Exception raised when there is an error with an Egg."""


@dataclass
class Egg:
    """An Habitica egg."""

    def __init__(self, name: str, *, quantity: int = 1):
        if name not in EggData.egg_names:
            raise EggException(f"{name} is not a valid egg name.")
        if quantity < 0:
            raise EggException(f"{quantity} is below 0.")

        self.name = name
        self.quantity = quantity

    def __eq__(self, other: Any) -> bool:
        """Return True if both eggs have the same name and same quantity."""
        if isinstance(other, Egg) is False:
            return False
        return self.name == other.name and self.quantity == other.quantity

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self.name}: {self.quantity})"

    def is_standard_egg(self) -> bool:
        """Return True if this is a drop egg. Else False"""
        return self.name in EggData.drop_egg_names

    def is_quest_egg(self) -> bool:
        """Return True if this is a quest egg. Else False"""
        return self.name in EggData.quest_egg_names

    def can_be_hatched_by(self, potion: HatchPotion) -> bool:
        """Return true if the specified potion can hatch this egg."""
        if potion.quantity == 0 or self.quantity == 0:
            return False
        if self.is_quest_egg():
            return potion.is_standard_hatch_potion()
        return True


@dataclass
class EggCollection:
    """Collection of eggs, backed by a Dict[str, Egg]."""

    eggs: Dict[str, int] = field(init=True, compare=False, default_factory=dict)
    __eggs: Dict[str, Egg] = field(init=False)

    def __post_init__(self):
        """Use the given eggs to create __eggs."""
        self.__eggs: Dict[str, Egg] = {
            name: Egg(name, quantity=quantity) for (name, quantity) in self.eggs.items()
        }

    def __len__(self) -> int:
        return len(self.__eggs)

    def __iter__(self):
        return iter(self.__eggs)

    def __getitem__(self, name: str) -> Egg:
        return self.__eggs[name]

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self.__eggs=})"

    def values(self):
        """Like dict.values but for an EggCollection."""
        for egg in self.__eggs.values():
            yield egg

    def remove_egg(self, egg: Egg) -> "EggCollection":
        """Remove a single eggs from the EggCollection.

        :param egg: egg to remove
        :return: The updated EggCollection
        """
        if self.__eggs.get(egg.name) is None:
            raise EggException(f"{egg.name} was not in the collection {self.__eggs}")
        if self.__eggs[egg.name].quantity == 0:
            raise EggException(f"We had 0 {egg.name} in the collection. Cannot remove any more.")

        self.__eggs[egg.name].quantity -= 1

        return self

    def get_standard_egg_collection(self) -> "EggCollection":
        """Return all the standard eggs in this collection."""
        return EggCollection({
            name: egg.quantity for (name, egg) in self.__eggs.items() if egg.is_standard_egg()
        })

    def get_quest_egg_collection(self) -> "EggCollection":
        """Return all the quest eggs in this collection."""
        return EggCollection({
            name: egg.quantity for (name, egg) in self.__eggs.items() if egg.is_quest_egg()
        })



#!/usr/bin/env python3
"""
The module with CLI code that handles the `hopla cast` command.
"""
import logging

import click
import requests

from hopla.cli.groupcmds.get_user import HabiticaUser
from hopla.hoplalib.cast.castcontroller import PostCastRequest
from hopla.hoplalib.cast.spellmodel import Spell, SpellData
from hopla.hoplalib.requests_helper import get_data_or_exit

log = logging.getLogger()


@click.command()
@click.argument("spell_name", type=click.Choice(SpellData.single_arg_spells))
def cast(spell_name):
    """Cast a spell.

    SPELL_NAME the name of the spell to cast

    Currently, there is only support for spells that are not cast on one
    single task.

    \f
    [APIDOCS](https://habitica.com/apidoc/#api-User-UserCast)
    :param spell_name:
    :return:
    """
    log.debug(f"hopla cast {spell_name=}")
    request = PostCastRequest(spell=Spell(spell_name))
    response: requests.Response = request.post_spell()

    json_data: dict = get_data_or_exit(response)
    mana: float = HabiticaUser(json_data["user"]).get_mp()
    click.echo(f"{spell_name} casted successfully: {mana:.1f} mana left.")



#!/usr/bin/env python3
from typing import List

import pytest

from hopla.hoplalib.hatchery.eggmodels import Egg, EggCollection
from hopla.hoplalib.hatchery.hatchalgorithms import HatchPlan, HatchPlanItem, HatchPlanMaker
from hopla.hoplalib.hatchery.hatchdata import EggData
from hopla.hoplalib.hatchery.hatchpotionmodels import HatchPotion, HatchPotionCollection
from hopla.hoplalib.zoo.foodmodels import FeedStatus
from hopla.hoplalib.zoo.petmodels import Pet


class TestHatchPlanItem:
    def test__repr__ok(self):
        egg_name, potion_name = "Wolf", "Aquatic"
        egg = Egg(egg_name, quantity=2)
        potion = HatchPotion(potion_name, quantity=3)
        plan_item = HatchPlanItem(egg=egg, potion=potion)

        result: str = repr(plan_item)

        assert result == f"HatchPlanItem({egg_name}: {potion_name})"

    @pytest.mark.parametrize("egg,potion", [
        (Egg("Wolf"), HatchPotion("Silver")),
        (Egg("Fox"), HatchPotion("Moonglow")),
        (Egg("Cactus"), HatchPotion("Watery")),
        (Egg("LionCub"), HatchPotion("Shade")),
        (Egg("PandaCub"), HatchPotion("Ruby")),
        (Egg("Whale"), HatchPotion("Base")),
        (Egg("Horse"), HatchPotion("Golden")),
        (Egg("Fox"), HatchPotion("Red"))
    ])
    def test__post_init_ok(self, egg: Egg, potion: HatchPotion):
        try:
            HatchPlanItem(egg=egg, potion=potion)
        except Exception:
            pytest.fail(f"Expected {egg.name} and {potion.name} to match.")

    def test__eq__should_eq_self_ok(self):
        egg_name, potion_name = "Snake", "Base"
        egg = Egg(egg_name)
        potion = HatchPotion(potion_name)
        item = HatchPlanItem(egg=egg, potion=potion)

        assert item == item

    def test__eq__should_ignore_quantity_differences_true(self):
        egg_name, potion_name = "Snake", "Base"

        item_self = HatchPlanItem(
            egg=Egg(egg_name, quantity=2),
            potion=HatchPotion(potion_name, quantity=3)
        )

        item_other = HatchPlanItem(
            egg=Egg(egg_name, quantity=12),
            potion=HatchPotion(potion_name, quantity=103)
        )

        assert item_self == item_other

    def test__eq__different_egg_name_false(self):
        potion_name = "Base"
        item_self = HatchPlanItem(egg=Egg("Snake"), potion=HatchPotion(potion_name))
        item_other = HatchPlanItem(egg=Egg("Horse"), potion=HatchPotion(potion_name))

        assert item_self != item_other

    def test__eq__different_potion_name_false(self):
        egg_name = "Horse"
        item_self = HatchPlanItem(egg=Egg(egg_name), potion=HatchPotion("Base"))
        item_other = HatchPlanItem(egg=Egg(egg_name), potion=HatchPotion("CottonCandyBlue"))

        assert item_self != item_other

    def test__eq__different_thing_false(self):
        item_self = HatchPlanItem(egg=Egg("Horse"), potion=HatchPotion("Base"))

        assert item_self != Egg("Snake")
        assert item_self != HatchPotion("Base")

    def test_format_item(self):
        egg_name = "Snake"
        potion_name = "Base"
        item = HatchPlanItem(egg=Egg(egg_name), potion=HatchPotion(potion_name))

        result: str = item.format_item()

        expected = f"A {egg_name} egg will be hatched by a {potion_name} potion.\n"
        assert result == expected

    @pytest.mark.parametrize("no_clash_pets", [
        Pet("Wolf-Bronze"),
        Pet("Fox-Base"),
        Pet("Wolf-Base", feed_status=FeedStatus(-1)),
        Pet("Wolf-Base", feed_status=FeedStatus(0))
    ])
    def test_can_hatch_with_pet_true(self, no_clash_pets: Pet):
        hatch_item = HatchPlanItem(egg=Egg("Wolf"), potion=HatchPotion("Base"))

        result = hatch_item.can_hatch_with_pet(no_clash_pets)

        assert result is True

    def test_can_hatch_with_pet_false(self):
        clash_pet = Pet("Wolf-Bronze")
        hatch_item = HatchPlanItem(egg=Egg("Wolf"), potion=HatchPotion("Bronze"))

        result = hatch_item.can_hatch_with_pet(clash_pet)

        assert result is False

    @pytest.mark.parametrize("hatch_item", [
        HatchPlanItem(Egg("Wolf"), HatchPotion("Bronze")),
        HatchPlanItem(Egg("Cactus"), HatchPotion("Amber")),
        HatchPlanItem(Egg("Horse"), HatchPotion("Shade"))
    ])
    def test_can_hatch_with_pets_no_pets_always_true(self, hatch_item: HatchPlanItem):
        result = hatch_item.can_hatch_with_pets([])
        assert result is True

    @pytest.mark.parametrize("hatch_item", [
        HatchPlanItem(Egg("Wolf"), HatchPotion("Bronze")),
        HatchPlanItem(Egg("Cactus"), HatchPotion("Amber")),
        HatchPlanItem(Egg("Horse"), HatchPotion("Shade")),
        HatchPlanItem(Egg("Wolf"), HatchPotion("Base")),
    ])
    def test_can_hatch_with_pets_no_clash_true(self, hatch_item: HatchPlanItem):
        no_clash_pets: List[Pet] = [
            Pet("Wolf-Base", feed_status=FeedStatus(0)),
            Pet("Cactus-Shade"),
            Pet("Seahorse-White"),
            Pet("Horse-Shade", feed_status=FeedStatus(-1))
        ]
        result = hatch_item.can_hatch_with_pets(no_clash_pets)
        assert result is True

    @pytest.mark.parametrize("hatch_item", [
        HatchPlanItem(Egg("Wolf"), HatchPotion("Bronze")),
        HatchPlanItem(Egg("Cactus"), HatchPotion("Amber")),
        HatchPlanItem(Egg("Horse"), HatchPotion("Shade")),
        HatchPlanItem(Egg("Wolf"), HatchPotion("Base")),
    ])
    def test_can_hatch_with_pets_clash_false(self, hatch_item: HatchPlanItem):
        clash_pets: List[Pet] = [
            Pet("Wolf-Bronze"), Pet("Cactus-Amber"), Pet("Horse-Shade"), Pet("Wolf-Base")
        ]
        result = hatch_item.can_hatch_with_pets(clash_pets)
        assert result is False


class TestHatchPlan:
    def test__init__(self):
        hatch_plan = HatchPlan()
        assert hatch_plan.plan == []

    def test__eq__empty(self):
        assert HatchPlan() == HatchPlan()

    def test__eq__false(self):
        plan = HatchPlan().add(egg=Egg("Whale"), potion=HatchPotion("Golden"))
        assert plan != HatchPlan()

    def test__repr__empty(self):
        hatch_plan = HatchPlan()

        result: str = repr(hatch_plan)

        assert result == "HatchPlan(empty)"

    def test__repr__not_empty(self):
        hatch_plan = HatchPlan()
        egg_name = "Horse"
        potion_name = "CottonCandyBlue"
        hatch_plan.add(egg=Egg(egg_name), potion=HatchPotion(potion_name))

        result: str = repr(hatch_plan)

        assert result == f"HatchPlan([HatchPlanItem({egg_name}: {potion_name})])"

    def test_format_plan(self):
        plan = HatchPlan() \
            .add(egg=Egg("Fox"), potion=HatchPotion("Dessert")) \
            .add(egg=Egg("Seahorse"), potion=HatchPotion("Desert"))

        result: str = plan.format_plan()

        expected: str = (
            "A Fox egg will be hatched by a Dessert potion.\n"
            "A Seahorse egg will be hatched by a Desert potion.\n"
        )

        assert result == expected

    def test_is_empty_true(self):
        plan = HatchPlan()
        assert plan.is_empty() is True
        assert len(plan) == 0

    def test_is_empty_false(self):
        plan = HatchPlan().add(egg=Egg("Egg"), potion=HatchPotion("Red"))
        assert plan.is_empty() is False
        assert len(plan) == 1

    def test_add_ok(self):
        hatch_plan = HatchPlan()
        egg = Egg("Seahorse")
        potion = HatchPotion("Skeleton")
        hatch_plan.add(egg=egg, potion=potion)

        assert len(hatch_plan) == 1
        assert HatchPlanItem(egg, potion) in hatch_plan

    def test__iter__(self):
        egg1, egg2 = Egg("Wolf"), Egg("Egg")
        ghost_potion = HatchPotion("Ghost")
        base_potion = HatchPotion("Base")
        hatch_plan = HatchPlan() \
            .add(egg=egg1, potion=ghost_potion) \
            .add(egg=egg2, potion=base_potion)

        iter_result = iter(hatch_plan)
        assert next(iter_result) == HatchPlanItem(egg=egg1, potion=ghost_potion)
        assert next(iter_result) == HatchPlanItem(egg=egg2, potion=base_potion)
        with pytest.raises(StopIteration):
            next(iter_result)


class TestHatchPlanMaker:
    def test__repr__(self):
        eggs = EggCollection({"Wolf": 1, "Whale": 2})
        potions = HatchPotionCollection({"Base": 1, "Ember": 2})
        plan_maker = HatchPlanMaker(egg_collection=eggs, hatch_potion_collection=potions, pets=[])

        result: str = repr(plan_maker)

        assert result == (
            f"HatchPlanMaker(\n"
            f"    {eggs},\n"
            f"    {potions},\n"
            f"    plan=HatchPlan(empty)\n"
            ")"
        )

    def test_make_plan_no_eggs(self):
        empty_eggs = EggCollection()
        potions = HatchPotionCollection({"Zombie": 1})

        plan_maker = HatchPlanMaker(
            egg_collection=empty_eggs,
            hatch_potion_collection=potions,
            pets=[]
        )

        result: HatchPlan = plan_maker.make_plan()

        expected_empty_plan = HatchPlan()
        assert result == expected_empty_plan

    def test_make_plan_no_hatch_potions(self):
        eggs = EggCollection({"Owl": 1})
        empty_potions = HatchPotionCollection()
        plan_maker = HatchPlanMaker(
            egg_collection=eggs,
            hatch_potion_collection=empty_potions,
            pets=[]
        )

        result: HatchPlan = plan_maker.make_plan()

        expected_empty_plan = HatchPlan()
        assert result == expected_empty_plan

    def test_make_plan_1potion_1egg_ok(self):
        eggs = EggCollection({"Owl": 1})
        potions = HatchPotionCollection({"Base": 1})
        plan_maker = HatchPlanMaker(
            egg_collection=eggs,
            hatch_potion_collection=potions,
            pets=[]
        )

        result: HatchPlan = plan_maker.make_plan()

        expected = HatchPlan().add(egg=Egg("Owl"), potion=HatchPotion("Base"))
        assert result == expected

    def test_make_plan_1magic_potion_1quest_egg_ok(self):
        eggs = EggCollection({"Owl": 1})
        magic_potions = HatchPotionCollection({"Thunderstorm": 1})
        plan_maker = HatchPlanMaker(
            egg_collection=eggs,
            hatch_potion_collection=magic_potions,
            pets=[]
        )

        result: HatchPlan = plan_maker.make_plan()

        expected_empty_plan = HatchPlan()
        assert result == expected_empty_plan

    def test_make_plan_1magic_potion_2normal_1quest_egg_ok(self):
        eggs = EggCollection({"Owl": 1, "Fox": 2})
        magic_potions = HatchPotionCollection({"Thunderstorm": 1})
        plan_maker = HatchPlanMaker(
            egg_collection=eggs,
            hatch_potion_collection=magic_potions,
            pets=[]
        )

        result: HatchPlan = plan_maker.make_plan()

        expected = HatchPlan().add(egg=Egg("Fox"),
                                   potion=HatchPotion("Thunderstorm"))
        assert result == expected

    def test_make_plan_more_potions_than_eggs(self):
        eggs = EggCollection({"Horse": 1, "Cactus": 2})
        potions = HatchPotionCollection(
            {"SolarSystem": 2, "Zombie": 1, "Base": 1, "Fluorite": 1, "Shade": 3}
        )
        plan_maker = HatchPlanMaker(
            egg_collection=eggs,
            hatch_potion_collection=potions,
            pets=[]
        )

        result: HatchPlan = plan_maker.make_plan()

        expected = HatchPlan() \
            .add(egg=Egg("Horse"), potion=HatchPotion("Zombie")) \
            .add(egg=Egg("Cactus"), potion=HatchPotion("SolarSystem")) \
            .add(egg=Egg("Cactus"), potion=HatchPotion("Base")) \
            # ^ Can't hatch same Cactus twice with same potion

        assert result == expected

    def test_make_plan_with_pets(self):
        eggs = EggCollection({"Horse": 1, "Cactus": 3})
        potions = HatchPotionCollection(
            {"SolarSystem": 2, "Zombie": 1, "Base": 1, "Fluorite": 1, "Shade": 3}
        )
        pets: List[Pet] = [
            # This pet blocks hatching another Cactus-SolarSystem
            Pet("Cactus-SolarSystem"),
            # No blocking by Horse-Zombie pet. -1 means pet is not available.
            Pet("Horse-Zombie", feed_status=FeedStatus(-1))
        ]
        plan_maker = HatchPlanMaker(
            egg_collection=eggs,
            hatch_potion_collection=potions,
            pets=pets
        )

        result: HatchPlan = plan_maker.make_plan()

        expected = HatchPlan() \
            .add(egg=Egg("Horse"), potion=HatchPotion("Zombie")) \
            .add(egg=Egg("Cactus"), potion=HatchPotion("Base")) \
            .add(egg=Egg("Cactus"), potion=HatchPotion("Fluorite")) \
            .add(egg=Egg("Cactus"), potion=HatchPotion("Shade"))

        assert result == expected

    def test_make_plan_with_lots_of_pets_and_few_potions_ok(self):
        # case emulating the bug from: https://github.com/melvio/hopla/issues/184
        # Here there are lots of pets, lots of eggs, and only 3 vampire potions.
        # There are also only 3 pets available for hatching with those vampire potions.

        # lots of Eggs
        eggs = EggCollection(eggs={egg_name: 20 for egg_name in EggData.egg_names})
        # 3 vampire potions
        potions = HatchPotionCollection(potions={"Vampire": 3})
        # lots of vampire pets, only 3 hatch-able
        hatchable1, hatchable2, hatchable3 = "FlyingPig", "LionCub", "PandaCub"
        pets: List[Pet] = [
            Pet("BearCub-Vampire", feed_status=FeedStatus(5)),
            Pet("Cactus-Vampire", feed_status=FeedStatus(5)),
            Pet("Dragon-Vampire", feed_status=FeedStatus(5)),
            Pet(hatchable1 + "-Vampire", feed_status=FeedStatus(-1)),
            Pet("Fox-Vampire", feed_status=FeedStatus(5)),
            Pet(hatchable2 + "-Vampire", feed_status=FeedStatus(-1)),
            Pet(hatchable3 + "-Vampire", feed_status=FeedStatus(-1)),
            Pet("TigerCub-Vampire", feed_status=FeedStatus(5)),
            Pet("Wolf-Vampire", feed_status=FeedStatus(5)),
        ]

        plan_maker = HatchPlanMaker(
            egg_collection=eggs,
            hatch_potion_collection=potions,
            pets=pets
        )

        plan: HatchPlan = plan_maker.make_plan()
        expected_plan = HatchPlan() \
            .add(egg=Egg(hatchable1), potion=HatchPotion("Vampire")) \
            .add(egg=Egg(hatchable2), potion=HatchPotion("Vampire")) \
            .add(egg=Egg(hatchable3), potion=HatchPotion("Vampire"))

        assert plan == expected_plan



#!/usr/bin/env python3
"""
A module with helper functions for the requests package.
"""
import logging
import sys
from typing import Any, NoReturn, Union

import requests
import click

from hopla.hoplalib.outputformatter import JsonFormatter

log = logging.getLogger()


def get_data_or_exit(api_response: requests.Response) -> Union[NoReturn, Any]:
    """Returns the "data" of a response if successful, else print error message and exit

    :param api_response:
    :return:
    """
    response_json = api_response.json()
    if response_json["success"]:
        return response_json["data"]

    __failed_to_get_data_so_exit(response_json=response_json,
                                 status_code=api_response.status_code)


def __failed_to_get_data_so_exit(response_json,
                                 status_code: int) -> NoReturn:
    log.debug(f"received: {response_json=}")
    click.echo(JsonFormatter(response_json).format_with_double_quotes())
    sys.exit(f"The habitica API call failed: {status_code=}")



#!/usr/bin/env python3
from unittest.mock import MagicMock, patch

import pytest
from click.testing import CliRunner, Result

from hopla.cli.hatch.quest_egg import quest_egg


class MockHatchResponse:
    def __init__(self, json):
        self._json = json

    def json(self):
        return self._json


class MockHatchRequester:
    def __init__(self, response_json):
        self._response_json = response_json

    def post_hatch_egg_request(self):
        return MockHatchResponse(self._response_json)


class TestHatchQuestEggCliCommand:
    # https://habitica.com/apidoc/#api-User-UserHatch

    @pytest.mark.parametrize("pet_name,potion_name", [
        ("Gryphon", "Base"),
        ("Hedgehog", "Red"),
    ])
    @patch("hopla.cli.groupcmds.hatch.HatchRequester")
    def test_standard_egg_ok(self, feed_requester_mock: MagicMock,
                             pet_name: str, potion_name: str):
        json: dict = {
            "success": True,
            "data": {"user.items": "way to much stuff here"},
            "message": "Your egg hatched! Visit your stable to equip your pet."
        }
        feed_requester_mock.return_value = MockHatchRequester(response_json=json)

        runner = CliRunner()
        result: Result = runner.invoke(quest_egg, [pet_name, potion_name])

        assert result.stdout == f"Successfully hatched a {pet_name}-{potion_name}.\n"
        assert result.exit_code == 0

    @pytest.mark.parametrize("error,message", [
        ("NotAuthorized", "You already have that pet. Try hatching a different combination!"),
        ("NotFound", "You're missing either that egg or that potion"),
    ])
    @patch("hopla.cli.groupcmds.hatch.HatchRequester")
    def test_standard_egg_fail(self, feed_requester_mock: MagicMock,
                               error: str, message: str):
        json: dict = {"success": False, "error": error, "message": message}
        feed_requester_mock.return_value = MockHatchRequester(response_json=json)

        runner = CliRunner()
        result: Result = runner.invoke(quest_egg, ["Frog", "White"])

        assert result.stdout == f"{error}: {message}\n"
        assert result.exit_code == 1



#!/usr/bin/env python3
"""
The module with CLI code that handles the `hopla buy enchanted-armoire` command.
"""
from dataclasses import dataclass
import logging
from typing import NoReturn, Optional, Union

import click
import requests

from hopla.hoplalib.requests_helper import get_data_or_exit
from hopla.hoplalib.http import HabiticaRequest, UrlBuilder
from hopla.hoplalib.outputformatter import JsonFormatter
from hopla.cli.groupcmds.get_user import HabiticaUserRequest, HabiticaUser
from hopla.hoplalib.throttling import ApiRequestThrottler

log = logging.getLogger()


@dataclass
class BuyEnchantedArmoireRequest(HabiticaRequest):
    """HabiticaRequest that buys from the enchanted armoire."""

    @property
    def url(self) -> str:
        """The habitica API buy url."""
        return UrlBuilder(path_extension="/user/buy-armoire").url

    def post_buy_request(self) -> requests.Response:
        """POST a buy request to the habitica API."""
        return requests.post(url=self.url, headers=self.default_headers)

    def post_buy_request_get_data_or_exit(self) -> Union[dict, NoReturn]:
        """POST a buy request and return the result, exit if the request failed.

        :return: If successful, the armoire content.
        """
        response: requests.Response = self.post_buy_request()
        # By default, we get way too much JSON info, so filter on "armoire".
        return get_data_or_exit(response)["armoire"]


def buy_from_enchanted_armoire_once():
    """buy once from the enchanted armoire"""
    buy_request = BuyEnchantedArmoireRequest()

    buy_data: dict = buy_request.post_buy_request_get_data_or_exit()

    enchanted_armoire_award = JsonFormatter(buy_data).format_with_double_quotes()
    click.echo(enchanted_armoire_award)


def times_until_poor(gp_budget: float) -> int:
    """Returns how many times you can buy from enchanted_armoire given some budget.

    \f
    :param gp_budget:
    :return:
    """
    enchanted_armoire_price = 100
    times: float = gp_budget / enchanted_armoire_price
    return int(times)  # int will round down positive floats for us


@click.command()
@click.option("--times", "-t", "requested_times",
              type=click.IntRange(min=0),
              metavar="[TIMES]",
              help="number of times to buy from the enchanted-armoire")
@click.option("--until-poor", "-u", "until_poor_flag",
              help="buy from enchanted-armoire until gp runs out",
              default=False, is_flag=True, show_default=True)
def enchanted_armoire(requested_times: int, until_poor_flag: bool):
    """Buy from the enchanted armoire

    TIMES - the number of times to buy from the enchanted armoire. Must be at least 0.
    When TIMES is omitted, 1 is the default. When TIMES is larger than the budget allows,
    --until-poor will be used instead. This behavior is likely to change in the
    future: <https://github.com/melvio/hopla/issues/100>

    If no options are specified, you buy once.

    \b
    Examples
    ----
    # buy once
    $ hopla buy enchanted-armoire

    \b
    # buy 3 times
    $ hopla buy enchanted-armoire -t3
    $ hopla buy enchanted-armoire -t 3
    $ hopla buy enchanted-armoire --times 3
    $ hopla buy enchanted-armoire --times=3

    \b
    # buy until you cannot afford the enchanted-armoire anymore
    $ hopla buy enchanted-armoire --until-poor
    $ hopla buy enchanted-armoire -u
    """
    log.debug(f"hopla buy enchanted-armoire times={requested_times}, until_poor={until_poor_flag}")

    user: HabiticaUser = HabiticaUserRequest().request_user_data_or_exit()
    times: int = get_buy_times_within_budget(user=user,
                                             until_poor_flag=until_poor_flag,
                                             requested_times=requested_times)

    buy_requests = [buy_from_enchanted_armoire_once for _ in range(times)]
    dispatcher = ApiRequestThrottler(buy_requests)
    dispatcher.execute_all_requests()


def get_buy_times_within_budget(*, user: HabiticaUser,
                                until_poor_flag: bool,
                                requested_times: Optional[int]) -> int:
    """Return how often we can buy, given the requested amount and our budget."""
    budget: float = user.get_gp()

    max_times = times_until_poor(budget)
    if until_poor_flag:
        wanted_times = max_times
    else:
        wanted_times = requested_times or 1

    times = min(max_times, wanted_times)
    click.echo(f"The current budget is {int(budget)}gp.\n"
               f"Buying: {times} times.", err=True)
    return times



#!/usr/bin/env python3
"""
Module with spell models and data.
"""
from dataclasses import dataclass
from typing import List, Optional
from uuid import UUID

from hopla.hoplalib.errors import YouFoundABugRewardError


class SpellData:
    """A class with data about spells.

    @see: habitica api content | jq. spells
    """

    warrior_spells_single_arg = ["defensiveStance", "valorousPresence", "intimidate"]
    mage_spells_single_arg = ["mpheal", "earth"]
    rogue_spells_single_arg = ["toolsOfTrade", "stealth"]
    healer_spells_single_arg = ["heal", "brightness", "protectAura", "healAll"]

    single_arg_spells: List[str] = (
            warrior_spells_single_arg
            + mage_spells_single_arg
            + rogue_spells_single_arg
            + healer_spells_single_arg
    )


@dataclass
class Spell:
    """A spell."""

    def __init__(self, spell_name: str,
                 *, target_id: UUID = None):
        if spell_name not in SpellData.single_arg_spells:
            raise YouFoundABugRewardError(f"{spell_name=} does not exist.")
        self.spell_name = spell_name
        self.target_id: Optional[UUID] = target_id
        """ not supported now """

    @property
    def class_name(self) -> str:
        """Get the Habitica class from the spell name.

        see: https://habitica.fandom.com/wiki/Class_System
        """
        if self.spell_name in SpellData.warrior_spells_single_arg:
            return "warrior"
        if self.spell_name in SpellData.mage_spells_single_arg:
            return "mage"
        if self.spell_name in SpellData.healer_spells_single_arg:
            return "healer"
        if self.spell_name in SpellData.rogue_spells_single_arg:
            return "rogue"

        raise YouFoundABugRewardError(f"{self.spell_name=} not implemented yet")
        # Mage: fireball="Burst of Flames", mpheal="Ethereal Surge", earth="Earthquake",
        #         frost="Chilling Frost"
        # Warrior: smash="Brutal Smash", defensiveStance="Defensive Stance",
        #          valorousPresence="Valorous Presence", intimidate="Intimidating Gaze"

        # Rogue: pickPocket="Pickpocket", backStab="Backstab", toolsOfTrade="Tools of the Trade",
        # stealth="Stealth"

        # Healer: heal="Healing Light", protectAura="Protective Aura",
        #          brightness="Searing Brightness",
        #         healAll="Blessing"
        # Transformation Items: snowball="Snowball", spookySparkles="Spooky Sparkles",
        #                       seafoam="Seafoam", shinySeed="Shiny Seed"



"""
The module with CLI code that handles the `hopla get` group command.
"""
import logging

import click

from hopla.hoplalib.user.usercontroller import HabiticaUserRequest
from hopla.hoplalib.user.usermodels import HabiticaUser

log = logging.getLogger()

pass_user = click.make_pass_decorator(HabiticaUser)


@click.group()
@click.pass_context
def get_user(ctx: click.Context) -> HabiticaUser:
    """
    GROUP for getting user information from Habitica.
    """
    log.debug("hopla get-user")
    user: HabiticaUser = HabiticaUserRequest().request_user_data_or_exit()
    ctx.obj = user
    return user
